/*
 * abiathar - main.c
 *
 * Copyright (C) 2019  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <gpio.h>
#include <tty.h>
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include <timer.h>
#include <rtu.h>
#include <wiegand.h>

#define GPIO_GREEN	0x40
#define GPIO_RED	0x41

#define GPIO_OUT0	0x16
#define GPIO_OUT1	0x15
#define GPIO_OUT2	0x14
#define GPIO_OUT3	0x13

#define GPIO_IN0	0x20
#define GPIO_IN1	0x21
#define GPIO_IN2	0x24
#define GPIO_IN3	0x28
#define GPIO_IN4	0x29
#define GPIO_IN5	0x2a
#define GPIO_IN6	0x2e
#define GPIO_IN7	0x2f

#define GPIO_W10	0x12
#define GPIO_W11	0x11
#define GPIO_W20	0x10
#define GPIO_W21	0x0f

#ifndef cpu_to_le32
#define cpu_to_le32(x)	(x)
#endif

#ifndef le32_to_cpu
#define le32_to_cpu(x)	(x)
#endif

#ifndef CHAR_BIT
#define CHAR_BIT		8
#endif

#ifndef BITS_PER_INT
#define BITS_PER_INT (sizeof(int) * CHAR_BIT)
#endif

static const unsigned int slave_addr = 1;

static unsigned int din;
static unsigned int dout;
static unsigned int leds;

enum {
	REG_VERSION = 0,
	REG_INPUTS = 1,
	REG_RELAYS = 2,
	REG_LEDS = 3, 
};

static int read_reg(unsigned int addr)
{
	if ((addr & 0xfff0) == 0x0010)
		return wiegand_read((addr & 8) ? 1 : 0, addr & 7);

	switch (addr) {
	case REG_VERSION:
		return 1;	/* version */
	case REG_INPUTS:
		return din;
	case REG_RELAYS:
		return dout;
	case REG_LEDS:
		return leds;
	default:
		if (addr < 32)
			return 0;
		return -EFAULT;
	}
}

static int write_reg(unsigned int value, unsigned int addr)
{
	if ((addr & 0xfff0) == 0x0010)
		return wiegand_write((addr & 8) ? 1 : 0, value, addr & 7);

	switch (addr) {
	case REG_RELAYS:
		dout = value & 0xf;
		break;
	case REG_LEDS:
		leds = value & 3;
		break;
	default:
		return -EFAULT;
	}
	return 0;
}

static unsigned int modbus_read16(const void *buf)
{
	return be16_to_cpu(get_unaligned((const u16 *) buf));
}

static void modbus_write16(unsigned int value, const void *buf)
{
	put_unaligned(cpu_to_be16(value), (u16 *) buf);
}

#define MBEX_NOSYS		0x01
#define MBEX_FAULT		0x02
#define MBEX_INVAL		0x03
#define MBEX_SLAVE_FAIL		0x04
#define MBEX_ACK		0x05
#define MBEX_BUSY		0x06
#define MBEX_MEMPARERR		0x08
#define MBEX_NOGW		0x0a
#define MBEX_GWERR		0x0b
static void modbus_rtu_recv(struct rtu *p, const unsigned char *frame,
			unsigned int len)
{
	unsigned int command;
	unsigned char rxbuf[256];
	unsigned char err = MBEX_NOSYS;
	unsigned int addr;
	unsigned int x;
	unsigned int num;
	unsigned int i;
	int ret;

	if (len < 2)
		return;
	if (frame[0] && frame[0] != slave_addr)
		return;
	command = frame[1];

	rxbuf[0] = frame[0];
	if (command == 3) {
		if (!frame[0])
			return;

		if (len != 6)
			goto err;

		addr = modbus_read16(frame + 2);
		num = modbus_read16(frame + 4);
		err = MBEX_INVAL;
		if (num < 1 || num > 0x7d)
			goto err;

		err = MBEX_FAULT;
		for (i = 0; i < num; i++) {
			ret = read_reg(addr + i);
			if (ret < 0)
				goto err;
			modbus_write16(ret, rxbuf + 3 + 2 * i);
		}
		rxbuf[1] = command;
		rxbuf[2] = 2 * num;
		rtu_xmit(p, rxbuf, 3 + 2 * num);
	} else if (command == 6) {
		if (len != 6)
			goto err;

		addr = modbus_read16(frame + 2);
		x = modbus_read16(frame + 4);
		ret = write_reg(x, addr);
		err = MBEX_FAULT;
		if (ret < 0)
			goto err;
		ret = read_reg(addr);
		if (ret < 0)
			goto err;

		rxbuf[1] = command;
		modbus_write16(addr, rxbuf + 2);
		modbus_write16(ret, rxbuf + 4);
		if (frame[0])
			rtu_xmit(p, rxbuf, 6);
	} else if (command == 0x10) {
		if (len < 7)
			goto err;
		err = MBEX_INVAL;
		addr = modbus_read16(frame + 2);
		num = modbus_read16(frame + 4);
		if (len != 7 + 2 * num)
			goto err;

		if (frame[6] != 2 * num)
			goto err;
		if (num < 1 || num > 0x7b)
			goto err;

		err = MBEX_FAULT;
		for (i = 0; i < num; i++) {
			x = modbus_read16(frame + 7 + 2 * i);
			ret = write_reg(x, addr + i);
			if (ret < 0)
				goto err;
		}
		rxbuf[1] = command;
		modbus_write16(addr, rxbuf + 2);
		modbus_write16(num, rxbuf + 4);
		if (frame[0])
			rtu_xmit(p, rxbuf, 6);
	} else {
		goto err;
	}

	return;

err:
	if (!frame[0])
		return;
	rxbuf[1] = command | 0x80;
	rxbuf[2] = err;
	rtu_xmit(p, rxbuf, 3);
}

static void din_update(void)
{
	static unsigned int x1; 
	unsigned int x = 0;

	x |= gpio_get_value(GPIO_IN0) << 0;
	x |= gpio_get_value(GPIO_IN1) << 1;
	x |= gpio_get_value(GPIO_IN2) << 2;
	x |= gpio_get_value(GPIO_IN3) << 3;
	x |= gpio_get_value(GPIO_IN4) << 4;
	x |= gpio_get_value(GPIO_IN5) << 5;
	x |= gpio_get_value(GPIO_IN6) << 6;
	x |= gpio_get_value(GPIO_IN7) << 7;
	x ^= 0xff;

	din = (din & x1) | ((din | x1) & x);
	x1 = x;
}

int main(void)
{
	struct tty_struct *tty0;
	struct rtu *rtu0;
	unsigned long t;

	gpio_direction_input(GPIO_IN0);
	gpio_direction_input(GPIO_IN1);
	gpio_direction_input(GPIO_IN2);
	gpio_direction_input(GPIO_IN3);
	gpio_direction_input(GPIO_IN4);
	gpio_direction_input(GPIO_IN5);
	gpio_direction_input(GPIO_IN6);
	gpio_direction_input(GPIO_IN7);

	gpio_direction_output(GPIO_RED, 0);
	gpio_direction_output(GPIO_GREEN, 1);

	gpio_direction_output(GPIO_OUT0, 0);
	gpio_direction_output(GPIO_OUT1, 0);
	gpio_direction_output(GPIO_OUT2, 0);
	gpio_direction_output(GPIO_OUT3, 0);

	wiegand_init(0, GPIO_W10, GPIO_W11);
	wiegand_init(1, GPIO_W20, GPIO_W21);

	local_irq_disable();
	tty0 = serial_init(0);
	rtu0 = rtu_init(tty0, modbus_rtu_recv);
	local_irq_enable();

	timer_init();

	t = jiffies;
	for (;;) {
		sleep_to(++t);

		din_update();
		wiegand_poll();

		gpio_set_value(GPIO_OUT0, (dout & 1) != 0);
		gpio_set_value(GPIO_OUT1, (dout & 2) != 0);
		gpio_set_value(GPIO_OUT2, (dout & 4) != 0);
		gpio_set_value(GPIO_OUT3, (dout & 8) != 0);

		gpio_set_value(GPIO_GREEN, (leds & 1) == 0);
		gpio_set_value(GPIO_RED, (leds & 2) == 0);

		rtu_poll(rtu0);
	}
	return 0;
}
