/*
 * grader-io - posix-serial.c
 *
 * Copyright (C) 2016  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <tty.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <stdio.h>
#include <signal.h>
#include <termios.h>

#define PORT_MAX	4

static struct tty_struct serial_tty[PORT_MAX];
static int serial_fd[PORT_MAX];
unsigned long serial_irq_time;

static int posix_tty_write(struct tty_struct *tty, const unsigned char *buf,
		int count)
{
	unsigned int port = (unsigned int)(tty - serial_tty);
	int ret;

	ret = write(serial_fd[port], buf, count);
	if (ret < 0)
		ret = -errno;
	return ret;
}

static int posix_tty_write_room(struct tty_struct *tty)
{
	return 128;
}

static const struct tty_operations posix_tty_ops = {
	.write		= posix_tty_write,
	.write_room	= posix_tty_write_room,
};

void sigio_handler(void)
{
	unsigned char buf[32];
	struct tty_struct *tty;
	int ret;
	unsigned int port;

	for (port = 0; port < PORT_MAX; port++) {
		if (serial_fd[port] <= 0)
			continue;

		ret = read(serial_fd[port], buf, sizeof(buf));
		if (ret < 0) {
			if (errno == EAGAIN || errno == EINTR || errno == EIO)
				continue;
			perror("read");
			exit(EXIT_FAILURE);
		}
		if (!ret)
			exit(EXIT_SUCCESS);

		tty = &serial_tty[port];

		if (!tty->ldisc_ops)
			continue;
		if (!tty->ldisc_ops->receive_buf)
			continue;
		tty->ldisc_ops->receive_buf(tty, buf, NULL, ret);
	}
}

void serial_poll(void)
{
	local_irq_disable();
	sigio_handler();
	local_irq_enable();
}

extern void signals_init(void);

struct tty_struct *serial_init(unsigned int port)
{
	struct tty_struct *tty;
	struct termios termios;
	int ret;
	int fd;

	if (port > PORT_MAX)
		return NULL;

	tty = &serial_tty[port];

	tty->ops = &posix_tty_ops;
	tty->ldisc_ops = NULL;

	signals_init();

	fd = posix_openpt(O_RDWR | O_NOCTTY);
	if (fd == -1) {
		perror("posix_openpt");
		exit(EXIT_FAILURE);
	}
	fprintf(stderr, "serial %d at '%s'\n", port, ptsname(fd));
	grantpt(fd);
	unlockpt(fd);

	tcgetattr(fd, &termios);
	termios.c_iflag &= ~(IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR
			| IGNCR | ICRNL | IXON);
	termios.c_oflag &= ~OPOST;
	termios.c_lflag &= ~(ECHO | ECHONL | ICANON | ISIG | IEXTEN);
	termios.c_cflag &= ~(CSIZE | PARENB);
	termios.c_cflag |= CS8;
	tcsetattr(fd, TCSANOW, &termios);

	ret = fcntl(fd, F_SETOWN, getpid());
	if (ret < 0)
		perror("fcntl");
	ret = fcntl(fd, F_SETFL, O_ASYNC | O_NONBLOCK);
	if (ret < 0)
		perror("fcntl");
	serial_fd[port] = fd;

	return tty;
}
