/*
 * abiathar - wiegand.c
 *
 * Wiegand protocol support
 *
 * Copyright (C) 2019  Krzysztof Mazur <krzysiek@podlesie.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <kernel.h>
#include <tty.h>
#include <timer.h>
#include <gpio.h>
#include <wiegand.h>

#define WIEGAND_CONTROL		0
#define WIEGAND_FORMAT_SHIFT	0
#define WIEGAND_FORMAT_MASK	0xff
#define WIEGAND_FORMAT_26	52
#define WIEGAND_FORMAT_34	56
#define WIEGAND_FORMAT_MAX	59

#define WIEGAND_FORMAT_P0_ODD	1
#define WIEGAND_FORMAT_P1_ODD	2

#define WIEGAND_PULSE		1
#define WIEGAND_BIT		2
#define WIEGAND_ACKSEQ		3
#define WIEGAND_DATA2		4
#define WIEGAND_DATA1		5
#define WIEGAND_DATA0		6
#define WIEGAND_TXSEQ		7

#define WIEGAND_MAX 2

#define WIEGAND_MTU 2

enum wiegand_state {
	WIEGAND_INIT = 0,
	WIEGAND_IDLE,
	WIEGAND_TX,
	WIEGAND_RX,
	WIEGAND_WAIT,
};

struct wiegand {
	u64 tx_buf;
	unsigned long next_jiffy;
	unsigned int tx_bits;

	u16 data[3];
	unsigned int tx_seq;
	unsigned int ack_seq;
	unsigned int format;

	unsigned int data0_gpio;
	unsigned int data1_gpio;

	unsigned int pulse_time;
	unsigned int bit_time;
};

static struct wiegand wiegand_table[WIEGAND_MAX];

#define le32_to_cpu(x)	(x)
#define cpu_to_le32(x)	(x)

static void __wiegand_poll(struct wiegand *p)
{
	if (!p->tx_bits)
		return;

	if (time_before(jiffies, p->next_jiffy))
		return;

	p->tx_bits--;
	if (p->tx_bits & 1) {
		unsigned int bit = (p->tx_buf & (1ULL << 47)) != 0ULL;

		gpio_set_value(p->data0_gpio, !bit);
		gpio_set_value(p->data1_gpio, bit);
		p->tx_buf <<= 1;
		p->next_jiffy += p->pulse_time;
	} else {
		gpio_set_value(p->data0_gpio, 0);
		gpio_set_value(p->data1_gpio, 0);
		if (p->bit_time > p->pulse_time)
			p->next_jiffy += p->bit_time - p->pulse_time;
		else
			p->next_jiffy++;

		if (!p->tx_bits)
			p->ack_seq = p->tx_seq;
	}
}

void wiegand_poll(void)
{
	struct wiegand *p;

	for (p = wiegand_table; p < wiegand_table + WIEGAND_MAX; p++)
		__wiegand_poll(p);
}

static int __wiegand_read(struct wiegand *p, unsigned int reg)
{
	if (reg >= 8)
		return -EFAULT;

	switch (reg) {
	case WIEGAND_CONTROL:
		return p->format << WIEGAND_FORMAT_SHIFT;
	case WIEGAND_PULSE:
		return 10 * p->pulse_time;
	case WIEGAND_BIT:
		return p->bit_time;
	case WIEGAND_ACKSEQ:
		return p->ack_seq;
	case WIEGAND_DATA2:
		return p->data[2];
	case WIEGAND_DATA1:
		return p->data[1];
	case WIEGAND_DATA0:
		return p->data[0];
	case WIEGAND_TXSEQ:
		return p->tx_seq;
	}
	return -EFAULT;
}

int wiegand_read(unsigned int dev, unsigned int reg)
{
	struct wiegand *p = wiegand_table + dev;

	if (dev >= WIEGAND_MAX)
		return -ENXIO;

	return __wiegand_read(p, reg);
}

#define parity(x) __builtin_parity(x)

static int wiegand_start_tx(struct wiegand *p)
{
	if (p->format <= 48) {
		p->tx_buf = (u64) p->data[0] << 32;
		p->tx_buf |= (u64) p->data[1] << 16;
		p->tx_buf |= (u64) p->data[2] << 0;
		p->tx_bits = 2 * p->format;
	} else if ((p->format & 0xfffc) == WIEGAND_FORMAT_26) {
		unsigned int data;
		unsigned int p0, p1;

		data = (p->data[1] & 0xff) << 16;
		data |= p->data[0];
		p0 = parity(data & 0xfff000);
		p1 = parity(data & 0x000fff);
		if (p->format & WIEGAND_FORMAT_P0_ODD)
			p0 = !p0;
		if (p->format & WIEGAND_FORMAT_P1_ODD)
			p1 = !p1;
		data |= p0 << 24;
		data <<= 1;
		data |= p1;

		p->tx_buf = ((u64) data) << (48 - 26);
		p->tx_bits = 2 * 26;
	} else if ((p->format & 0xfffc) == WIEGAND_FORMAT_34) {
		unsigned int data;
		u64 data64;
		unsigned int p0, p1;

		data = p->data[1] << 16;
		data |= p->data[0];
		p0 = parity(p->data[1]);
		p1 = parity(p->data[0]);
		if (p->format & WIEGAND_FORMAT_P0_ODD)
			p0 = !p0;
		if (p->format & WIEGAND_FORMAT_P1_ODD)
			p1 = !p1;
		data64 = p0;
		data64 <<= 32;
		data64 |= data;
		data64 <<= 1;
		data64 |= p1;
		data64 <<= 48 - 34;

		p->tx_buf = data64;
		p->tx_bits = 2 * 34;
	}
	p->next_jiffy = jiffies + 1;
	return 0;
}

static int __wiegand_write(struct wiegand *p, unsigned int x, unsigned int reg)
{
	if (reg >= 8)
		return -EFAULT;

	switch (reg) {
	case WIEGAND_CONTROL:
		p->format = (x >> WIEGAND_FORMAT_SHIFT) & WIEGAND_FORMAT_MASK;
		if (p->format > WIEGAND_FORMAT_MAX)
			p->format = 0;
		if (p->format >= 49 && p->format < 52)
			p->format = 0;
		break;
	case WIEGAND_PULSE:
		p->pulse_time = (x + 9) / 10;
		break;
	case WIEGAND_BIT:
		p->bit_time = x;
		break;
	case WIEGAND_ACKSEQ:
		break;
	case WIEGAND_DATA2:
		p->data[2] = x;
		break;
	case WIEGAND_DATA1:
		p->data[1] = x;
		break;
	case WIEGAND_DATA0:
		p->data[0] = x;
		break;
	case WIEGAND_TXSEQ:
		if (x == ((p->tx_seq + 1) & 0xffff)) {
			p->tx_seq = x;
			wiegand_start_tx(p);
		}
		break;
	}
	return 0;
}

int wiegand_write(unsigned int dev, unsigned int value, unsigned int reg)
{
	struct wiegand *p = wiegand_table + dev;

	if (dev >= WIEGAND_MAX)
		return -ENXIO;

	return __wiegand_write(p, value, reg);
}

int wiegand_init(unsigned int dev, unsigned int data0, unsigned int data1)
{
	struct wiegand *p = wiegand_table + dev;

	if (dev >= WIEGAND_MAX)
		return -ENXIO;

	gpio_direction_output(data0, 0);
	gpio_direction_output(data1, 0);
	p->data0_gpio = data0;
	p->data1_gpio = data1;
	p->format = WIEGAND_FORMAT_26 | WIEGAND_FORMAT_P1_ODD;
	p->pulse_time = 1;
	p->bit_time = 10;
	return 0;
}
